# coding=utf-8

from django.db import models
from django.utils import six


class LlToEarth(models.Expression):
    """
    Returns the location of a point on the surface of the Earth given its
    latitude and longitude in degrees.
    """
    template = 'll_to_earth(%(params)s)'

    def __init__(self, params, output_field=None):
        if output_field is None:
            output_field = models.fields.Field()
        self.params = params
        super(LlToEarth, self).__init__(output_field=output_field)

    def resolve_expression(self, query=None, allow_joins=True, reuse=None, summarize=False, for_save=False):
        """Setup any data here, this method will be called before final SQL is generated"""
        c = self.copy()
        c.is_summary = summarize
        c.for_save = for_save
        final_points = []
        for i, p in enumerate(self.params):
            try:
                float(p)
            except:
                _, source, _, join_list, last = query.setup_joins(
                    six.text_type(p).split('__'), query.model._meta, query.get_initial_alias())
                target, alias, _ = query.trim_joins(source, join_list, last)
                final_points.append("%s.%s" % (alias, target[0].get_attname_column()[1]))
            else:
                final_points.append(six.text_type(p))
            c.params = final_points
        return c

    def as_sql(self, compiler, connection):
        """Returns ll_to_earth(field,field)"""
        return self.template % {'params': ','.join(self.params)}, []


class EarthDistance(models.Expression):
    """
    This PostgreSQL function returns the great circle distance between two points on the surface of the
    earth.
    """
    template = 'earth_distance(%(expressions)s)'

    def __init__(self, expressions, output_field=None):
        if output_field is None:
            output_field = models.fields.Field()
        self.expressions = expressions
        super(EarthDistance, self).__init__(output_field=output_field)

    def resolve_expression(self, query=None, allow_joins=True, reuse=None, summarize=False, for_save=False):
        """Prepare SQL from inner funcions (ll_to_earth or any other)"""
        c = self.copy()
        c.is_summary = summarize
        c.for_save = for_save
        for pos, expression in enumerate(self.expressions):
            c.expressions[pos] = expression.resolve_expression(query, allow_joins, reuse, summarize)
        return c

    def as_sql(self, compiler, connection):
        sql_expressions, sql_params = [], []
        for expression in self.expressions:
            sql, params = compiler.compile(expression)
            sql_expressions.append(sql)
            sql_params.extend(params)
        return self.template % {'expressions': ','.join(sql_expressions)}, sql_params


class EarthDistanceQuerySet(models.QuerySet):
    def in_distance(self, distance, fields, points, annotate='_ed_distance'):
        """Filter rows inside a circunference of radius distance `distance`

            :param distance: max distance to allow
            :param fields: `tuple` with the fields to filter (latitude, longitude)
            :param points: center of the circunference (latitude, longitude)
            :param annotate: name where the distance will be annotated

        """
        clone = self._clone()
        return clone.annotate(
            **{annotate: EarthDistance([
                LlToEarth(fields), LlToEarth(points)])
            }).filter(**{'{0}__lte'.format(annotate): distance})




# coding=utf-8

from django.test import TestCase

from django_earthdistance.models import EarthDistance, LlToEarth

from .models import TestModel


class EarthdistanceTest(TestCase):
    def setUp(self):
        self.obj_1 = TestModel.objects.create(lat=51.490857, lon=-0.15071)
        self.obj_2 = TestModel.objects.create(lat=39.74829, lon=-1.23854)
        self.obj_3 = TestModel.objects.create(lat=39.46487, lon=-1.23854)

    def test_in_distance(self):
        objects = TestModel.objects.in_distance(
            1500, ('lat', 'lon'), (float(51.490857), float(-0.15071)))
        self.assertEqual(objects.count(), 1)
        self.assertEqual(objects[0].pk, self.obj_1.pk)
        self.assertTrue(hasattr(objects[0], '_ed_distance'))
        objects = TestModel.objects.in_distance(
            100000, ('lat', 'lon'), (float(39.49087), float(-1.15071)))
        self.assertEqual(objects.count(), 2)
        self.assertTrue(
            all(o.pk in [self.obj_2.pk, self.obj_3.pk] for o in objects))

    def test_annotate(self):
        obj = TestModel.objects.annotate(
            distance=EarthDistance([
                LlToEarth(('lat', 'lon')),
                LlToEarth((float(39.49087), float(-1.15071)))
            ])).get(id=self.obj_1.id)
        self.assertEqual(int(obj.distance), 1338082)




# def register_geometric_types(connection):
#     import objects

#     for objectname in objects.object_names:
#         obj_class = getattr(objects, objectname)
#         obj_class.register_cast(connection)
#         obj_class.register_adapter()


# from djorm_core.models import connection_handler
# connection_handler.attach_handler(register_geometric_types, vendor="postgresql", unique=True)


# coding=utf-8

from django.db import models

from django_earthdistance.models import EarthDistanceQuerySet

class TestModel(models.Model):
    lat = models.FloatField()
    lon = models.FloatField()

    objects = EarthDistanceQuerySet.as_manager()

